package com.ukixa.barcode.readersample

import android.Manifest
import android.content.pm.ActivityInfo
import android.content.pm.PackageManager
import android.media.AudioAttributes
import android.media.SoundPool
import android.os.Bundle
import android.util.Log
import android.util.Size
import android.view.WindowManager
import android.widget.TextView
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.camera.core.*
import androidx.camera.lifecycle.ProcessCameraProvider
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import com.ukixa.barcode.BarcodeReader
import com.ukixa.barcode.BarcodeType
import com.ukixa.barcode.readersample.databinding.ActivityMainBinding
import java.nio.ByteBuffer
import java.util.concurrent.ExecutorService
import java.util.concurrent.Executors

typealias BarcodeListener = (text: String) -> Unit

class MainActivity : AppCompatActivity() {
    private lateinit var viewBinding: ActivityMainBinding
    private lateinit var cameraExecutor: ExecutorService
    private lateinit var soundPool: SoundPool
    private var readerBeepSoundID = 0

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        viewBinding = ActivityMainBinding.inflate(layoutInflater)
        setContentView(viewBinding.root)

        window.addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON)
        // サンプルコードを単純化するため端末は縦向きで固定とする。
        requestedOrientation = ActivityInfo.SCREEN_ORIENTATION_PORTRAIT

        val audioAttributes = AudioAttributes.Builder()
            .setUsage(AudioAttributes.USAGE_NOTIFICATION).build()
        soundPool = SoundPool.Builder()
            .setMaxStreams(1).setAudioAttributes(audioAttributes).build()
        readerBeepSoundID = soundPool.load(this, R.raw.beep, 1)

        val barcodeReader = BarcodeReader.getInstance()
        Log.d(
            TAG,
            "version string: ${BarcodeReader.getVersion()}"
        ) // versionプロパティにはライブラリの著作権情報とバージョン番号（文字列による）が格納されている。
        Log.d(
            TAG,
            "version number: ${BarcodeReader.getVersionNumber().toString(16)}"
        ) // versionNumberプロパティにはライブラリのバージョン番号（数値による）が格納されている。
        Log.d(TAG, "license: ${barcodeReader.license}") // licenseプロパティにはライブラリの使用許諾番号が格納されている。

        // バーコードリーダーオブジェクトの内部状態をリセットする。
        // ビデオフレームの処理中（画像解析メソッドanalyze(image: ImageProxy)内で）は、resetしないこと。
        barcodeReader.reset()

        // typesプロパティに読み取りたいバーコードの種類を設定する。
        // BarcodeTypeに定義された定数をビット単位ORして指定する。
        // 例えば、NW-7だけを読み取り対象としたい場合は types = BarcodeType.NW7 とする。
        // EAN-8とEAN-13の2種類が対象なら types = BarcodeType.EAN8 or BarcodeType.EAN13 とすればよい。
        barcodeReader.types = BarcodeType.NW7 or
                BarcodeType.ITF or
                BarcodeType.CODE39 or
                BarcodeType.CODE128 or
                BarcodeType.EAN13 or
                BarcodeType.EAN8 or
                BarcodeType.UPCE or
                BarcodeType.GS1DATABAR

        // rotationプロパティにイメージバッファの回転量を設定する。
        // 端末本体が縦向きであっても、デフォルトではカメラからのビデオフレームは横向きに送られてくる。
        // rotation指定により、デコード前にライブラリの内部処理でイメージを回転させる。
        barcodeReader.rotation = 3 // 270° = 90° x 3 回転
        // ビデオフレームを横向きのまま使用する場合、あるいは、
        // 前もってビデオフレームが縦向きになるように設定している場合は、rotation = 0; （回転なし）とする。

        // Request camera permissions
        if (allPermissionsGranted()) {
            startCamera()
        } else {
            ActivityCompat.requestPermissions(
                this, REQUIRED_PERMISSIONS, REQUEST_CODE_PERMISSIONS
            )
        }

        cameraExecutor = Executors.newSingleThreadExecutor()
    }

    private fun startCamera() {
        val cameraProviderFuture = ProcessCameraProvider.getInstance(this)

        cameraProviderFuture.addListener({
            // Used to bind the lifecycle of cameras to the lifecycle owner
            val cameraProvider: ProcessCameraProvider = cameraProviderFuture.get()

            // Preview
            val preview = Preview.Builder()
                .setTargetAspectRatio(AspectRatio.RATIO_16_9)
                .build()
                .also {
                    it.setSurfaceProvider(viewBinding.viewFinder.surfaceProvider)
                }

            val imageAnalyzer = ImageAnalysis.Builder()
                .setTargetResolution(Size(720, 1280)) // ビデオカメラの解像度（画素数）を指定する。
                .build()
                .also {
                    it.setAnalyzer(cameraExecutor, BarcodeAnalyzer { text ->
                        soundPool.play(readerBeepSoundID, 1.0f, 1.0f, 0, 0, 1.0f)
                        val barcodeView: TextView = findViewById(R.id.barcode_view)
                        barcodeView.text = text
                    })
                }

            // Select back camera as a default
            val cameraSelector = CameraSelector.DEFAULT_BACK_CAMERA

            try {
                // Unbind use cases before rebinding
                cameraProvider.unbindAll()

                // Bind use cases to camera
                cameraProvider.bindToLifecycle(
                    this, cameraSelector, preview, imageAnalyzer
                )

            } catch (exc: Exception) {
                Log.e(TAG, "Use case binding failed", exc)
            }

        }, ContextCompat.getMainExecutor(this))
    }

    private fun allPermissionsGranted() = REQUIRED_PERMISSIONS.all {
        ContextCompat.checkSelfPermission(
            baseContext, it
        ) == PackageManager.PERMISSION_GRANTED
    }

    override fun onDestroy() {
        super.onDestroy()
        cameraExecutor.shutdown()
    }

    override fun onRequestPermissionsResult(
        requestCode: Int, permissions: Array<String>, grantResults:
        IntArray
    ) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)
        if (requestCode == REQUEST_CODE_PERMISSIONS) {
            if (allPermissionsGranted()) {
                startCamera()
            } else {
                Toast.makeText(
                    this,
                    "Permissions not granted by the user.",
                    Toast.LENGTH_SHORT
                ).show()
                finish()
            }
        }
    }

    companion object {
        private const val TAG = "KXABR"
        private const val REQUEST_CODE_PERMISSIONS = 10
        private val REQUIRED_PERMISSIONS =
            mutableListOf(
                Manifest.permission.CAMERA,
            ).toTypedArray()
    }

    private class BarcodeAnalyzer(private val listener: BarcodeListener) : ImageAnalysis.Analyzer {
        private var previousScanTime = System.currentTimeMillis()
        private var previousScanText = ""

        private fun ByteBuffer.toByteArray(): ByteArray {
            rewind()    // Rewind the buffer to zero
            val data = ByteArray(remaining())
            get(data)   // Copy the buffer into a byte array
            return data // Return the byte array
        }

        override fun analyze(imageProxy: ImageProxy) {
            // イメージバッファを取得する。
            val buffer = imageProxy.planes[0].buffer
            //Log.d(TAG, "image.width: ${image.width}, image.height: ${image.height}")
            val barcodeReader = BarcodeReader.getInstance()
            // イメージバッファからバーコードを読み取る。
            val text = barcodeReader.decode(buffer, imageProxy.width, imageProxy.height)
            // decode() メソッドが正常に完了した場合、空でない文字列が返される。
            // 完了していない場合は空文字列となる。
            // 返り値と同じ文字列がtextプロパティにも保持される。
            if (text.isNotEmpty()) {
                previousScanText = text
                val currentTime = System.currentTimeMillis()
                var d = currentTime - previousScanTime

                if (text != previousScanText) {
                    // 同じバーコードの値が連続出力されてデバッグ表示が見にくくなるのを防ぐ。
                    d = 2600
                }

                if (d >= 2600) {
                    previousScanTime = currentTime
                    previousScanText = text

                    // typeプロパティには読み取ったバーコードの種類が格納されている。
                    val barcodeType =
                    when (barcodeReader.type) {
                        BarcodeType.NW7 -> "NW-7"
                        BarcodeType.ITF -> "ITF"
                        BarcodeType.CODE39 -> "CODE39"
                        BarcodeType.CODE128 -> "CODE128"
                        BarcodeType.EAN13 -> "EAN-13"
                        BarcodeType.EAN8 -> "EAN-8"
                        BarcodeType.UPCE -> "UPC-E"
                        BarcodeType.GS1DATABAR -> "GS1 DataBar"
                        else -> "UNKNOWN"
                    }

                    Log.d(TAG, "type: $barcodeType")
                    Log.d(TAG, text)

                    listener(text) //検出されたバーコード文字列を返す。
                }
            }

            imageProxy.close()
        }
    }
}
